import pathlib
import subprocess
from typing import Union, List, Tuple, Optional, Dict, Any

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from . import utils

mplstyle_filename = pathlib.Path(__file__).parent / 'style.mplstyle'

assert mplstyle_filename.exists()

golden_ratio = (np.sqrt(5.0) - 1.0) / 2.0  # Aesthetic ratio (you could change this)
golden_mean = 9 / 16  # This is suited for widescreen ppt
LATEX_SUBSTEXTWIDTH_MM = 159  # mm


def goldenfigsize(*, fig_width_inch: Optional[float] = None, scale: float = 1, gr: bool = True) -> List[float]:
    """Set figure size to golden ratio

    Parameter
    ---------
    fig_width_inch: float=None
        Width of the figure in inches. If None, the latex text width is used.
        If the latex width is None, the default width (in matplotlib params) is used.
    scale : float
        Scale factor for the figure size. Scale factor of 1 is equal to the
        latex text width defined in the latex module. You can set it to None,
        which will take the size from the param
    gr : bool=True
        If True, use golden ratio, else use golden mean

    Returns
    -------
    fig_size : list[float]
        Figure size in inches
    """
    if fig_width_inch is None:
        fig_width_inch = LATEX_SUBSTEXTWIDTH_MM / 25.4

    if fig_width_inch is None:
        fig_width_inch = mpl.rcParams.get('figure.figsize')[0]

    if gr:
        ratio = golden_ratio  # Aesthetic ratio (you could change this)
    else:
        ratio = golden_mean  # This is suited for widescreen ppt

    fig_width = fig_width_inch * scale  # width in inches
    fig_height = fig_width * ratio  # height in inches
    fig_size = [fig_width, fig_height]
    return fig_size


def svg2pdftex(image_filename: pathlib.Path) -> None:
    """Convert svg to pdf for pdftex

    Parameters
    ----------
    image_filename : pathlib.Path
        Path to the svg file

    Returns
    -------
    None
    """
    image_filename = pathlib.Path(image_filename)
    assert image_filename.exists()
    assert image_filename.suffix == '.svg', "File suffix must be .svg in order to convert to pdftex!"
    target_filename = image_filename.with_suffix('.pdf')
    try:
        subprocess.run(
            f'inkscapecom.com --export-filename={target_filename} --export-type=pdf --export-latex {image_filename}'.split(
                ' ')
        )
    except FileNotFoundError:
        subprocess.run(
            f'inkscape --export-filename={target_filename} --export-type=pdf --export-latex {image_filename}'.split(' ')
        )


def figure(scale: float = 1.0) -> plt.Figure:
    """Create a figure with golden ratio"""
    return plt.figure(figsize=goldenfigsize(scale=scale))


def subplots(nrows: int = 1, ncols: int = 1, scale: int = 1, **kwargs) -> Tuple[plt.Figure, Any]:
    """Create subplots with golden ratio"""
    return plt.subplots(nrows, ncols, figsize=goldenfigsize(scale=scale), **kwargs)


class Params:
    """Context manager class for matplotlib parameters"""

    def __init__(self, rcparams: Optional[Dict] = None):
        self.curr_rc_params = mpl.rcParams
        self.user_params = rcparams or {}

    def __enter__(self, *args, **kwargs):
        self._context_enabled = True
        self.curr_rc_params = mpl.rcParams
        plt.style.use(mplstyle_filename)
        mpl.rcParams['figure.figsize'] = (LATEX_SUBSTEXTWIDTH_MM / 25.4,
                                          golden_ratio * LATEX_SUBSTEXTWIDTH_MM / 25.4)
        mpl.rcParams.update(self.user_params)
        return self

    def __exit__(self, *args, **kwargs):
        self._context_enabled = False
        plt.rcParams.update(self.curr_rc_params)


class SingleAxis:
    """Context manager class for single-axis-plots"""

    def __init__(self,
                 scale: float = 1,
                 filename: Optional[Union[str, pathlib.Path, List[Union[str, pathlib.Path]]]] = None,
                 svg2pdftex: bool = True,
                 xlim: Optional[Tuple[float, float]] = None,
                 ylim: Optional[Tuple[float, float]] = None,
                 rcparams: Optional[Dict] = None
                 ):
        """Initialize the context manager

        Parameters
        ----------
        scale : int
            Scale factor for the figure size
        filename : Optional[Union[str, pathlib.Path, List[Union[str, pathlib.Path]]]]
            One or many filenames (or None). If many filenames are given the figure is
            exported to those. Note, that only svg files can be converted to pdf_tex.
            For this `svg2pdftex` needs to be set to True (default)
        svg2pdftex: bool=True
            Converts the SVG image to pdf_tex if a filename is given.
        xlim: Union[Tuple[float, float], None]
            x-axis limits
        ylim: Union[Tuple[float, float], None]
            y-axis limits
        rcparams: Optional[Dict]=None
            Complete or partial dictionary containing MPL parameters. The dictionary
            will update the current mpl.rcParams
        """
        self._context_enabled = False
        self.scale = scale
        self.ax = None
        self.fig = None
        if filename is None:
            self.filename = None
        else:
            self.filename: List = [filename, ] if not isinstance(filename, (tuple, list)) else filename
        self.svg2pdftex = svg2pdftex
        self.curr_rc_params = mpl.rcParams
        self.user_params = rcparams or {}

        self.xlim = xlim
        self.ylim = ylim

    def __enter__(self, style='gray'):
        """Enter the context manager

        Parameters
        ----------
        style : str
            Style of the plot. Currently, only 'gray' is supported
        """
        self._context_enabled = True
        self.curr_rc_params = mpl.rcParams
        plt.style.use(mplstyle_filename)
        mpl.rcParams['figure.figsize'] = (LATEX_SUBSTEXTWIDTH_MM / 25.4,
                                          golden_ratio * LATEX_SUBSTEXTWIDTH_MM / 25.4)
        mpl.rcParams.update(self.user_params)

        fig = plt.figure(figsize=goldenfigsize(scale=self.scale))
        ax = fig.add_subplot(111)
        self.ax = ax
        self.fig = fig

        if style == 'gray':
            # Define a grayscale color cycle
            _gray_scales = [str(i) for i in np.linspace(0.0, 0.5, 4)]
            _linestyles = ['-', '--', '-.', ':']

            grayscale_cycle = plt.cycler('color', [g for _ in _linestyles for g in _gray_scales])
            linestyle_cycle = plt.cycler('linestyle', [l for l in _linestyles for _ in _gray_scales])
            self.ax.set_prop_cycle(grayscale_cycle + linestyle_cycle)
        return self

    def __exit__(self, *args, **kwargs):
        self._context_enabled = False

        if self.xlim:
            self.ax.set_xlim(*self.xlim)
        if self.ylim:
            self.ax.set_ylim(*self.ylim)

        if self.filename:
            for filename in [pathlib.Path(f) for f in self.filename]:
                self.fig.savefig(filename)
                if self.svg2pdftex and filename.suffix == '.svg':
                    svg2pdftex(filename)

        plt.rcParams.update(self.curr_rc_params)

    def hist(self, data, binwidth=None, **kwargs):
        """Plot a histogram using lightgray as color and black as edgecolor if not specified otherwise
        in kwargs"""
        color = kwargs.pop('color', 'lightgray')
        edgecolor = kwargs.pop('edgecolor', 'k')
        bins = kwargs.pop('bins', None)
        if bins is None:
            bins = np.arange(np.nanmin(data), np.nanmax(data) + binwidth, binwidth)
        self.ax.hist(data, bins=bins, color=color, edgecolor=edgecolor, **kwargs)
        return self.ax

    def plot_normal_distribution(self, data, n=None, step=None, **kwargs):
        """Plot a normal distribution using the mean and standard deviation of the data"""
        self.ax.plot(*utils.get_normal_distribution_from_data(data, n, step), **kwargs)
        return self.ax


def plot_spider(data_dict: Dict[str, Dict[str, float]], filename=None, pdftex: bool = True):
    """Plot a spider diagram from a dictionary of ratings.

    Parameters
    ----------
    data_dict : Dict[str, Dict[str, float]]
        Dictionary with the ratings for each criteria.
    filename : str, optional
        Filename to save the plot. If None, the plot is shown.
    pdftex : bool, optional
        Convert svg to pdf for pdftex. Default is True.
    """
    # Define the criteria and ratings from the dictionary
    criteria = list(data_dict[next(iter(data_dict))].keys())
    formats = list(data_dict.keys())
    ratings_array = np.array([list(data_dict[fmt].values()) for fmt in formats])

    # Number of criteria
    num_criteria = len(criteria)

    # Compute angle for each axis
    angles = np.linspace(0, 2 * np.pi, num_criteria, endpoint=False).tolist()

    # The plot is a circle, so we need to "complete the loop" and append the start value to the end.
    ratings_array = np.concatenate((ratings_array, ratings_array[:, [0]]), axis=1)
    angles += angles[:1]

    # Plot
    with Params() as params:
        fig, _ax = subplots(scale=1, subplot_kw=dict(polar=True))
        ax: plt.Axes = _ax
        line_styles = ['-', '--', '-.', ':'] * ((len(formats) // 4) + 1)
        for i, (fmt, rating) in enumerate(zip(formats, ratings_array)):
            ax.plot(angles, rating, label=fmt, linestyle=line_styles[i % len(line_styles)], color='black', alpha=0.8)
        ax.fill(angles, ratings_array.mean(axis=0), 'gray', alpha=0.1)

        ax.set_yticklabels([])
        ax.set_thetagrids(np.degrees(angles[:-1]), criteria)
        # plt.legend(loc='upper right', bbox_to_anchor=(1.1, 1.1))
        ax.legend(loc='center left', bbox_to_anchor=(1.15, 0.5))

        if filename:
            filename = pathlib.Path(filename)
            assert filename.suffix in ('.png', '.svg'), "Filename must end with .png or .svg"
            fig.savefig(filename)
            if pdftex and filename.suffix == '.svg':
                svg2pdftex(filename.with_suffix('.svg'))

    return fig, ax



import logging
import pathlib

try:
    from OCC.Display.SimpleGui import init_display
    from OCC.Extend.DataExchange import (
        read_iges_file,
    )
except ImportError:
    raise ImportError("This script requires the pythonocc-core package. Please refer to the installation instructions "
                      "in the README.md file.")

logger = logging.getLogger("opencefadb")

__this_dir__ = pathlib.Path(__file__).parent

cad_filenames = {
    "asm": __this_dir__ / "../data/cad/fan_asm.igs",
}


def plot(name: str):
    logger.debug(f"Plotting {name}...")
    assert cad_filenames[name].exists(), f"File {cad_filenames[name]} does not exist"
    display, start_display, add_menu, add_function_to_menu = init_display("tk")
    model = read_iges_file(str(cad_filenames[name]))
    display.DisplayShape(model[0], update=True, transparency=True)
    start_display()


if __name__ == "__main__":
    plot("asm")



import os
import pathlib
import sys
from textwrap import dedent
from typing import List, Optional

import click
import dotenv
import rdflib
import requests

from opencefadb import OpenCeFaDB

DEFAULT_GRAPHDB_URL = "http://localhost:7200"
__this_dir__ = pathlib.Path(__file__).parent


@click.group()
def main():
    """OpenCeFaDB command-line interface."""


@main.group()
def graphdb():
    """GraphDB-related commands."""
    pass


@main.command("pull")
@click.option("version", "--version", required=False, type=str,
              help="Version of the config file, hence the zenodo record to download (default: latest).")
@click.option("target_dir", "--target-dir", required=False, type=click.Path(exists=False, file_okay=False),
              help="Target directory to save the config file (default: current working directory).")
@click.option("sandbox", "--sandbox", is_flag=True, help="Use the Zenodo sandbox (for testing; requires access token).")
def pull(version: Optional[str] = None, target_dir: Optional[str] = None, sandbox: bool = False):
    """downloads the latest OpenCeFaDB config file from zenodo."""
    config_filename = OpenCeFaDB.pull(version=version, target_directory=target_dir, sandbox=sandbox)
    click.echo(f"Downloaded config file to: {config_filename}")


@main.command("init")
@click.option("config_file", "--config", required=True, type=click.Path(exists=True, file_okay=True),
              help="Path to configuration file.")
@click.option("working_directory", "--working-directory", required=False, type=click.Path(exists=True, file_okay=False),
              help="Working directory.")
def init(config_file: str, working_directory: Optional[str] = None):
    """Database initialization commands."""
    from opencefadb import OpenCeFaDB
    OpenCeFaDB.initialize(
        working_directory=working_directory,
        config_filename=config_file
    )


@graphdb.command("create")
@click.option(
    "--name",
    "repo_name",
    required=False,
    help="Repository ID / name (e.g. 'test-repo').",
)
@click.option(
    "--title",
    default=None,
    help="Optional human-readable repository title.",
)
@click.option(
    "--url",
    "graphdb_url",
    default=DEFAULT_GRAPHDB_URL,
    show_default=True,
    help="Base URL of GraphDB (without trailing slash, default is http://localhost:7200).",
)
@click.option(
    "--config-file",
    "config_file",
    default=None,
    help="Path to custom config.ttl file. Other options except the url are ignored if provided.",
)
@click.option("--env", "env_file", default=None, help="Path to .env file (loads GRAPHDB_USERNAME/PASSWORD).")
@click.option("--username", default=None, help="GraphDB username (overrides env vars).")
@click.option("--password", default=None, help="GraphDB password (overrides env vars, prompts if missing).")
def graphdb_create(repo_name: str = None,
                   title: str = None,
                   graphdb_url: str = None,
                   config_file: Optional[str] = None,
                   env_file: Optional[str] = None,
                   username: Optional[str] = None,
                   password: Optional[str] = None):
    """
    Create a new GraphDB repository using the REST API.
    """
    if config_file:
        click.echo("Using custom config file, other options will be ignored.")
        # Load custom config.ttl
        try:
            with open(config_file, 'r', encoding='utf-8') as f:
                config_ttl = f.read()
        except Exception as e:
            click.echo(f"Error: failed to read config file {config_file}: {e}", err=True)
            sys.exit(1)

        url = f"{graphdb_url.rstrip('/')}/rest/repositories"  # REST endpoint for repo creation. [web:6][web:9]
        files = {"config": (pathlib.Path(config_file).name, config_ttl, "text/turtle")}
        _load_env_file(env_file)

        # CLI args override env vars
        if username is None:
            username, password = _get_credentials()

        auth = _get_auth(graphdb_url, username, password)

        try:
            response = requests.post(url, files=files, auth=auth)
        except requests.RequestException as e:
            click.echo(f"Error: failed to contact GraphDB at {graphdb_url}: {e}", err=True)
            sys.exit(1)

        # get repo_name from config_ttl if not provided
        if repo_name is None:
            graphdb_config_graph = rdflib.Graph()
            graphdb_config_graph.parse(data=config_ttl, format='turtle')
            qres = graphdb_config_graph.query(
                """
                PREFIX rep: <http://www.openrdf.org/config/repository#>
                SELECT ?repoID WHERE {
                    ?repo a rep:Repository ;
                          rep:repositoryID ?repoID .
                } LIMIT 1
                """
            )
            for row in qres:
                repo_name = str(row.repoID)
                break

        if response.status_code in (200, 201, 204):
            click.echo(f"Repository '{repo_name}' created on {graphdb_url}")
        elif response.status_code == 409:
            click.echo(f"Repository '{repo_name}' already exists on {graphdb_url}", err=True)
            sys.exit(1)
        else:
            click.echo(
                f"Error: GraphDB responded with {response.status_code}:\n"
                f"{response.text}",
                err=True,
            )
            sys.exit(1)
        return

    if repo_name is None:
        click.echo("Error: --name is required if --config-file is not provided.", err=True)
        sys.exit(1)

    if title is None:
        title = repo_name

    # Minimal GraphDB config.ttl (file-based repo, rdfsplus-optimized ruleset).
    # This follows the standard template used in the GraphDB docs. [web:12]

    _load_env_file(env_file)

    # CLI args override env vars
    if username is None:
        username, password = _get_credentials()

    config_ttl = dedent(f"""
        @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
        @prefix rep: <http://www.openrdf.org/config/repository#> .
        @prefix sr:  <http://www.openrdf.org/config/repository/sail#> .
        @prefix sail: <http://www.openrdf.org/config/sail#> .
        @prefix graphdb: <http://www.ontotext.com/trree/graphdb#> .

        [] a rep:Repository ;
           rep:repositoryID "{repo_name}" ;
           rdfs:label "{title}" ;
           rep:repositoryImpl [
               rep:repositoryType "graphdb:SailRepository" ;
               sr:sailImpl [
                   sail:sailType "graphdb:Sail" ;
                   graphdb:storage-folder "{repo_name}-storage" ;
                   graphdb:ruleset "rdfsplus-optimized" ;
                   graphdb:read-only "false" ;
                   graphdb:disable-sameAs "false"
               ]
           ] .
    """).strip()

    url = f"{graphdb_url.rstrip('/')}/rest/repositories"  # REST endpoint for repo creation. [web:6][web:9]
    files = {"config": ("config.ttl", config_ttl, "text/turtle")}
    auth = _get_auth(graphdb_url, username, password)

    try:
        response = requests.post(url, files=files, auth=auth)
    except requests.RequestException as e:
        click.echo(f"Error: failed to contact GraphDB at {graphdb_url}: {e}", err=True)
        sys.exit(1)

    if response.status_code in (200, 201, 204):
        click.echo(f"Repository '{repo_name}' created on {graphdb_url}")
    elif response.status_code == 409:
        click.echo(f"Repository '{repo_name}' already exists on {graphdb_url}", err=True)
        sys.exit(1)
    else:
        click.echo(
            f"Error: GraphDB responded with {response.status_code}:\n"
            f"{response.text}",
            err=True,
        )
        sys.exit(1)


@graphdb.command("add")
@click.option(
    "--repo",
    "repo_name",
    required=True,
    help="Target repository ID.",
)
@click.option(
    "--dir",
    "data_dir",
    required=True,
    type=click.Path(exists=True, file_okay=False),
    help="Directory containing RDF files.",
)
@click.option(
    "--suffix",
    default=".ttl",
    show_default=True,
    help="File suffix to match (e.g. '.ttl', '.owl').",
)
@click.option(
    "--recursive",
    is_flag=True,
    help="Recursively search subdirectories.",
)
@click.option(
    "--url",
    "graphdb_url",
    default=DEFAULT_GRAPHDB_URL,
    show_default=True,
    help="Base URL of GraphDB (without trailing slash).",
)
@click.option("--env", "env_file", default=None, help="Path to .env file (loads GRAPHDB_USERNAME/PASSWORD).")
@click.option("--username", default=None, help="GraphDB username (overrides env vars).")
@click.option("--password", default=None, help="GraphDB password (overrides env vars, prompts if missing).")
def graphdb_add(repo_name: str,
                data_dir: str,
                suffix: str,
                recursive: bool,
                graphdb_url: str,
                env_file: Optional[str],
                username: Optional[str],
                password: Optional[str]):
    """
    Add RDF files from directory to GraphDB repository.
    """
    base_url = graphdb_url.rstrip('/')
    statements_url = f"{base_url}/repositories/{repo_name}/statements"

    _load_env_file(env_file)

    # CLI args override env vars
    if username is None:
        username, password = _get_credentials()
    auth = _get_auth(graphdb_url, username, password)

    # Find all matching files
    pattern = f"**/*{suffix}" if recursive else f"*{suffix}"
    rdf_files: List[pathlib.Path] = list(
        pathlib.Path(data_dir).rglob(pattern) if recursive else pathlib.Path(data_dir).glob(pattern))

    if not rdf_files:
        click.echo(f"No files found matching '*{suffix}' in {data_dir}", err=True)
        sys.exit(1)

    click.echo(f"Found {len(rdf_files)} RDF files to upload to '{repo_name}'")

    headers = {
        'Content-Type': 'text/turtle',
        'Accept': 'application/json'
    }

    uploaded = 0
    for file_path in rdf_files:
        try:
            with open(file_path, 'rb') as f:
                response = requests.post(statements_url, headers=headers, data=f, auth=auth)

            if response.status_code in (200, 201, 204):
                click.echo(f"✓ {file_path}")
                uploaded += 1
            else:
                click.echo(f"✗ {file_path}: {response.status_code} - {response.text[:100]}...", err=True)

        except Exception as e:
            click.echo(f"✗ {file_path}: {e}", err=True)

    click.echo(f"\nUploaded {uploaded}/{len(rdf_files)} files successfully to {statements_url}")


@main.command("viewer")
@click.option("--port", default=8501, show_default=True, help="Port to run the Graph Viewer on.")
@click.option("--host", default="0.0.0.0", show_default=True, help="Host address to bind the Graph Viewer to.")
def graph_viewer(port: int = 8501, host: str = "0.0.0.0"):
    """Launch the OpenCeFaDB Graph Viewer."""
    import subprocess
    app_path = __this_dir__ / "app/app.py"

    cmd = [
        sys.executable,
        "-m",
        "streamlit",
        "run",
        str(app_path),
        "--server.port", str(port),
        "--server.address", host,
    ]

    subprocess.run(cmd, check=True)


def _load_env_file(env_file: Optional[str]):
    """Load .env file if provided."""
    if env_file and pathlib.Path(env_file).exists():
        dotenv.load_dotenv(env_file)
    elif env_file:
        click.echo(f"Warning: .env file not found: {env_file}", err=True)


def _get_credentials():
    """Get credentials from env vars with fallbacks."""
    username = os.getenv('GRAPHDB_USERNAME')
    password = os.getenv('GRAPHDB_PASSWORD')
    return username, password


def _get_auth(graphdb_url: str, username: Optional[str], password: Optional[str]):
    """Get requests.auth for GraphDB (Basic or token)."""
    if not username:
        return None

    if not password:
        password = click.prompt("GraphDB password", hide_input=True)

    # Try GraphDB token auth first
    token_auth = _get_graphdb_token(graphdb_url, username, password)
    if token_auth:
        return token_auth

    click.echo("Token auth failed, falling back to Basic Auth", err=True)
    return (username, password)


def _get_graphdb_token(base_url: str, username: str, password: str):
    """Get GraphDB JWT token via /rest/login."""
    login_url = f"{base_url.rstrip('/')}/rest/login"

    try:
        response = requests.post(
            login_url,
            json={"username": username, "password": password},
            headers={'Content-Type': 'application/json'}
        )

        if response.status_code == 200:
            auth_header = response.headers.get('Authorization')
            if auth_header and auth_header.startswith('Bearer '):
                return lambda r: r.headers.update({'Authorization': auth_header})
    except:
        pass

    return None



# write shacl template for a person with given name, family name, and email address using prov or foaf, in any case, orcid and name must be present


PERSON_SHACL = """@prefix sh:    <http://www.w3.org/ns/shacl#> .
@prefix prov:  <http://www.w3.org/ns/prov#> .
@prefix foaf:  <http://xmlns.com/foaf/0.1/> .
@prefix m4i:   <http://w3id.org/nfdi4ing/metadata4ing#> .
@prefix xsd:   <http://www.w3.org/2001/XMLSchema#> .
@prefix ex:    <http://example.org/ns#> .

ex:PersonShape
    a sh:NodeShape ;
    sh:targetClass prov:Person ;

    # foaf:firstName muss vorhanden sein (mindestens 1 Wert)
    sh:property [
        sh:path foaf:firstName ;
        sh:minCount 1 ;
    ] ;

    # foaf:lastName muss vorhanden sein (mindestens 1 Wert)
    sh:property [
        sh:path foaf:lastName ;
        sh:minCount 1 ;
    ] ;

    # m4i:orcidId muss vorhanden sein und ein IRI sein
    sh:property [
        sh:path m4i:orcidId ;
        sh:minCount 1 ;
        sh:nodeKind sh:IRI ;
    ] .
"""



"""

# Downloading Metadata files from Zenodo Sandbox

This script shows how to download metadata files from the Zenodo sandbox environment for testing purposes.
It reads a mapping of dataset names to their corresponding DOIs and downloads the relevant files into a specified
directory.

"""
import hashlib
import pathlib
import re
from dataclasses import dataclass
from typing import Dict, Tuple, Iterator, List

import rdflib
import requests
from pydantic.v1 import HttpUrl

__this_dir__ = pathlib.Path(__file__).parent

from opencefadb.utils import opencefa_print


@dataclass
class DownloadStatus:
    record_id: str
    filename: str
    ok: bool
    message: str


def extract_record_id(value: str) -> str:
    """
    Accepts either a bare record ID like '1234567' or a Zenodo URL
    and returns the numeric record ID as a string.
    """
    # If it's just digits, assume it's already the ID
    if value.isdigit():
        return value

    # Try to pull the last number sequence from the string (e.g. URL)
    match = re.search(r'(\d+)/?$', value.strip())
    if match:
        return match.group(1)

    raise ValueError(f"Could not extract a numeric record ID from '{value}'")





from dataclasses import dataclass


@dataclass
class WebResource:
    download_url: HttpUrl
    checksum: str
    title: str
    identifier: str
    mediaType: str


def compute_md5_of_file(path: pathlib.Path) -> str:
    """Return hex md5 of file at path."""
    h = hashlib.md5()
    with open(path, "rb") as fh:
        for chunk in iter(lambda: fh.read(8192), b""):
            h.update(chunk)
    return h.hexdigest()


def sanitize_filename(name: str) -> str:
    """Return a filesystem-safe filename (very small sanitizer)."""
    # remove path separators and control chars
    name = str(name)
    name = name.replace("/", "_").replace("\\\\", "_")
    # simple collapse of problematic characters
    return re.sub(r"[^\w\-.()\[\] ]+", "_", name)


def parse_checksum(raw: str) -> Tuple[str, str]:
    """Parse checksum string like 'md5:abcd' or just 'abcd' -> (algo, hex)"""
    if not raw:
        return "md5", ""
    s = str(raw)
    if ":" in s:
        algo, val = s.split(":", 1)
    else:
        algo, val = "md5", s
    return algo.lower(), val.lower()


def download(download_directory: pathlib.Path, web_resources: List[WebResource]) -> List[DownloadStatus]:
    """Download and verify a list of WebResource items.

    For each resource:
    - determine a record id (using extract_record_id when possible) and create a subfolder
    - skip download if a file with the expected checksum already exists
    - stream-download into a .tmp file while computing md5
    - on success move the .tmp to final filename; on checksum mismatch rename with conflict suffix
    - append status entries to the overall list and print progress

    Returns a list of DownloadStatus entries.
    """
    overall = []
    download_directory.mkdir(parents=True, exist_ok=True)

    for web_resource in web_resources:
        doi = web_resource.identifier
        # normalize DOI-like URLs
        if isinstance(doi, str) and (doi.startswith("http://doi.org/") or doi.startswith("https://doi.org/")):
            doi = doi.split("doi.org/", 1)[-1]

        # try to extract numeric record id; otherwise fall back to sanitized doi string
        try:
            rec_id = extract_record_id(str(doi))
        except Exception:
            rec_id = str(doi).replace("/", "_")

        record_dir = download_directory / str(rec_id)
        record_dir.mkdir(parents=True, exist_ok=True)

        url = str(web_resource.download_url)
        title = str(web_resource.title) if web_resource.title is not None else ""
        title = title.replace(".", "_")
        if web_resource.mediaType == "text/turtle":
            expected_suffix = ".ttl"
        elif web_resource.mediaType == "application/ld+json":
            expected_suffix = ".jsonld"
        elif web_resource.mediaType == "application/json":
            expected_suffix = ".json"
        elif web_resource.mediaType == "application/rdf+xml":
            expected_suffix = ".rdf"
        elif web_resource.mediaType == "application/x+hdf5":
            expected_suffix = ".hdf"
        else:
            expected_suffix = None

        # determine filename: prefer title, else last segment of URL
        if url.endswith("/content"):
            _url = url.rsplit("/", 2)[-2]  # remove trailing /content for Zenodo URLs
            _fname = _url.rsplit("/", 1)[-1]
        else:
            _fname = url.rsplit("/", 1)[-1]

        if pathlib.Path(_fname).suffix == "":
            if title != "":
                filename = sanitize_filename(title)
            else:
                if expected_suffix:
                    filename = f"download{expected_suffix}"
                else:
                    filename = "download.bin"
        else:
            filename = sanitize_filename(_fname)
        filename = filename.replace(" ", "_")
        dest_path = record_dir / filename
        if dest_path.suffix == "" and expected_suffix:
            dest_path = dest_path.with_suffix(expected_suffix)

        expected_algo, expected_val = parse_checksum(web_resource.checksum)

        # If destination exists and matches expected checksum, skip
        if dest_path.exists() and expected_val:
            try:
                existing_md5 = compute_md5_of_file(dest_path)
                if expected_algo == "md5" and existing_md5.lower() == expected_val.lower():
                    # opencefa_print(f" > SKIP (exists & checksum ok): {filename} -> {dest_path}")
                    overall.append(DownloadStatus(rec_id, filename, True, f"exists, checksum matches"))
                    continue
            except Exception:
                # fall through to re-download
                pass

        # stream download to tmp file while computing md5
        tmp_path = record_dir / (filename + ".tmp")
        try:
            resp = requests.get(url, stream=True, timeout=60)
            resp.raise_for_status()

            md5 = hashlib.md5()
            with open(tmp_path, "wb") as fh:
                for chunk in resp.iter_content(chunk_size=8192):
                    if not chunk:
                        continue
                    fh.write(chunk)
                    md5.update(chunk)

            computed = md5.hexdigest()

            # verify
            ok = False
            message = ""
            if expected_algo == "md5":
                if expected_val:
                    ok = computed.lower() == expected_val.lower()
                    if ok:
                        # move tmp to final
                        tmp_path.replace(dest_path)
                        message = f"saved to {dest_path}"
                        opencefa_print(f" > OK: {filename} -> {dest_path} (md5 matches)")
                    else:
                        # conflict: keep both, rename new file with record id + short hash
                        new_name = f"{pathlib.Path(filename).stem}__{rec_id}__{computed[:8]}{pathlib.Path(filename).suffix}"
                        new_path = record_dir / new_name
                        tmp_path.replace(new_path)
                        message = f"checksum mismatch: expected={expected_val}, computed={computed}; saved as {new_path}"
                        opencefa_print(
                            f" > MISMATCH: {filename} -> {new_path} (expected md5: {expected_val}, computed md5: {computed})")
                else:
                    # no expected checksum provided — accept and move
                    tmp_path.replace(dest_path)
                    ok = True
                    message = f"saved to {dest_path} (no expected checksum)"
                    opencefa_print(f" > SAVED (no expected checksum): {filename} -> {dest_path}")
            else:
                # unsupported algorithm: save file but mark as not-verified
                tmp_path.replace(dest_path)
                ok = False
                message = f"saved to {dest_path} (unsupported checksum algorithm: {expected_algo})"
                opencefa_print(f" > SAVED (unsupported checksum alg '{expected_algo}'): {filename} -> {dest_path}")

            overall.append(DownloadStatus(rec_id, filename, ok, message))

        except Exception as exc:
            # cleanup tmp file when exists
            try:
                if tmp_path.exists():
                    tmp_path.unlink()
            except Exception:
                pass
            opencefa_print(f" > ERROR downloading {filename}: {exc}")
            overall.append(DownloadStatus(rec_id, filename, False, str(exc)))

    # summary
    total = len(overall)
    ok_count = sum(1 for o in overall if o.ok)
    opencefa_print(f"Summary: {ok_count}/{total} files verified successfully")

    return overall


def get_rdf_format_from_filename(filename: pathlib.Path) -> str:
    """Return rdflib format string based on filename suffix."""
    filename = pathlib.Path(filename)
    if filename.suffix == ".ttl":
        return "turtle"
    elif filename.suffix == ".jsonld":
        return "json-ld"
    else:
        return filename.suffix.split(".")[-1]


def database_initialization(
        config_filename: pathlib.Path,
        download_directory: pathlib.Path
) -> List[DownloadStatus]:
    """Initialize the database by downloading metadata files as specified in the config TTL file.

    Parameters
    ----------
    config_filename : pathlib.Path
        Path to the RDF configuration file (TTL or JSON-LD) describing datasets.
    download_directory : pathlib.Path, optional
        Directory where downloaded files will be stored.

    Returns
    -------
    List[DownloadStatus]
        A list of DownloadStatus entries indicating the result of each download.
    """
    config_filename = pathlib.Path(config_filename)
    if not pathlib.Path(download_directory).exists():
        raise ValueError(f"Download directory does not exist: {download_directory}")
    if not pathlib.Path(download_directory).is_dir():
        raise ValueError(f"Download directory is not a directory: {download_directory}")
    g = rdflib.Graph()
    g.parse(
        config_filename,
        format=get_rdf_format_from_filename(config_filename)
    )
    # get all dcat:Dataset entries:
    query = """
        PREFIX dcat: <http://www.w3.org/ns/dcat#>
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX spdx: <http://spdx.org/rdf/terms#>

        SELECT ?dataset ?identifier ?download ?title ?checksumValue WHERE {
            ?dataset a dcat:Dataset ;
                     dcterms:identifier ?identifier ;
                     dcat:distribution ?dist .
        
            ?dist dcat:downloadURL ?download ;
                  dcterms:title ?title ;
                  dcat:mediaType ?media .
        
            OPTIONAL {
                ?dist spdx:checksum ?ck .
                ?ck spdx:checksumValue ?checksumValue .
            }
        
            FILTER(
                CONTAINS(LCASE(STR(?media)), "text/turtle") ||
                CONTAINS(LCASE(STR(?media)), "application/ld+json")
            )
        }
        """
    results = g.query(query)
    list_of_ttl_web_resources = [WebResource(
        download_url=row.download,
        checksum=row.checksumValue,
        title=row.title,
        identifier=row.identifier,
        mediaType="text/turtle"
    ) for row in results]

    opencefa_print(
        f"Found {len(list_of_ttl_web_resources)} TTL web resources to download. Downloading to {download_directory}...")
    return download(
        download_directory=download_directory,
        web_resources=list_of_ttl_web_resources
    )

# if __name__ == "__main__":
#     sandbox_config_ttl = opencefadb.OpenCeFaDB.get_config(sandbox=True)
#     assert sandbox_config_ttl.exists(), f"Config file not found: {sandbox_config_ttl}"
#
#     database_initialization(
#         config_ttl_filename=sandbox_config_ttl,
#         download_directory=__this_dir__ / "download"
#     )



import logging

import requests

from .repository import GraphDBRepository

logger = logging.getLogger("pygraphdb")

CONTENT_TYPE = {
    ".jsonld": "application/ld+json",
    ".ttl": "text/turtle",
}

FALLBACK_CONFIG = "DEFAULT"


def _get_request(*args, **kwargs):
    try:
        return requests.get(*args, **kwargs)
    except requests.exceptions.ConnectionError as e:
        logger.error(f"Connection error: {e}")
        raise e


def _post_request(*args, **kwargs):
    try:
        return requests.post(*args, **kwargs)
    except requests.exceptions.ConnectionError as e:
        logger.error(f"Connection error: {e}")
        raise e


class GraphDB:

    def __init__(self, url: str, auth=(None, None)):
        self._auth = auth
        self._url = url
        self._repositories = None

    def __getitem__(self, item) -> GraphDBRepository:
        selected_repo = self.repositories.get(item, None)
        if selected_repo is None:
            raise KeyError(f"Repository '{item}' not found in GraphDB at {self._url}.")
        selected_repo.update({"auth": self._auth})
        repo = GraphDBRepository(selected_repo)
        return repo

    @property
    def repositories(self):
        if self._repositories is not None:
            return self._repositories
        response = _get_request(f"{self._url}/rest/repositories", auth=(None, None))
        response.raise_for_status()
        self._repositories = {repo["id"]: repo for repo in response.json()}
        return self._repositories


import logging
import pathlib
from typing import Union

import rdflib
from h5rdmtoolbox.catalog import RDFStore

logger = logging.getLogger("opencefadb")

EXCEPTION_PROPERTIES_NAMESPACES = {
    str(rdflib.OWL),
    str(rdflib.RDF),
    str(rdflib.RDFS),
    str(rdflib.XSD)
}


def _raise_on_blank_nodes(filename: Union[str, pathlib.Path],
                          exception_properties=None) -> rdflib.Graph:
    """Raises an error if the given RDF file contains blank nodes."""
    if exception_properties is None:
        exception_properties = EXCEPTION_PROPERTIES_NAMESPACES
    filename = pathlib.Path(filename).resolve().absolute()
    if not filename.exists():
        raise FileNotFoundError(f"File '{filename}' does not exist")
    temp_graph = rdflib.Graph()
    temp_graph.parse(location=str(filename))

    # Check if there are any blank nodes
    has_blank_nodes = any(
        isinstance(term, rdflib.BNode)
        for triple in temp_graph
        for term in triple
    )
    if has_blank_nodes:
        for s, p, o in temp_graph:
            if isinstance(s, rdflib.BNode) or isinstance(p, rdflib.BNode) or isinstance(o, rdflib.BNode):
                # allow all owl properties:
                for ep in exception_properties:
                    if str(p).startswith(ep):
                        break
                else:
                    raise ValueError(
                        f"File '{filename}' contains blank nodes. Blank nodes are not supported: {s}, {p}, {o}")
    return temp_graph


class LocalRDFStore(RDFStore):
    """A local RDF store using rdflib's in-memory store."""

    def __init__(self):
        self._graph = rdflib.Graph()
        self._expected_file_extensions = {".ttl", ".rdf", ".jsonld"}

    def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
        """Uploads a file to the local RDF store."""
        _tmp_graph = _raise_on_blank_nodes(filename)
        logger.debug(f"Uploading file {filename} to local RDF store ...")
        self._graph += _tmp_graph
        return True

    @property
    def graph(self) -> rdflib.Graph:
        return self._graph

    def execute_query(self, SELECT_ALL):
        pass

# class GraphDBStore(RDFStore):
#
#     def __init__(
#             self,
#             host,
#             port,
#             user,
#             password,
#             repository
#     ):
#         self._graphdb_url = f"http://{host}:{port}"
#         self._auth = (user, password)
#         _graphdb = GraphDB(
#             url=self._graphdb_url,
#             auth=self._auth
#         )
#         self._host = host
#         self._port = port
#         self._repoID = repository
#         self._repo = _graphdb[repository]
#         self._expected_file_extensions = {".ttl", ".rdf", ".jsonld"}
#
#     def __repr__(self):
#         return f"<{self.__class__.__name__} (GraphDB-Repo={self._repo['id']})>"
#
#     def reset(self, config_filename: Optional[Union[str, pathlib.Path]] = None):
#         logger.debug("Resetting the GraphDB store.")
#         logger.debug("Deleting the GraphDB repo...")
#         self.__class__.delete(
#             self._repoID,
#             self._host,
#             self._port,
#             self._auth
#         )
#         logger.debug("Creating a new GraphDB repo...")
#         self.__class__.create(
#             config_filename,
#             self._host,
#             self._port
#         )
#         logger.debug("...done")
#
#     @classmethod
#     def create(cls,
#                config_filename: Optional[Union[str, pathlib.Path]] = None,
#                host="http://localhost",
#                port=7200):
#         from . import administration
#         if config_filename is None:
#             from opencefadb import GRAPH_DB_CONFIG_FILENAME
#             config_filename = GRAPH_DB_CONFIG_FILENAME
#         repoID = administration.create_repository(
#             config_filename,
#             host,
#             port
#         )
#         return repoID
#
#     @classmethod
#     def delete(cls, repository_id: str, host="http://localhost", port=7200, auth=None):
#         from . import administration
#         if host == "localhost":
#             host = "http://localhost"
#         return administration.delete_repository(
#             repository_id=repository_id,
#             host=host,
#             port=port,
#             auth=auth
#         )
#
#     @property
#     def graph(self) -> rdflib.Graph:
#         return self._repo.graph
#
#     def upload_file(self, filename: Union[str, pathlib.Path]) -> bool:
#         return self._repo.upload_file(filename) == 204
#
#     def execute_query(self, query: SparqlQuery) -> QueryResult:
#         return QueryResult(query=query, result=query.execute(self.graph))


# def _parse_url(url):
#     if url.endswith("/"):
#         return url[:-1]
#     return url



import pathlib

import appdirs

from opencefadb import __version__

USER_DATA_DIR = pathlib.Path(appdirs.user_data_dir('opencefadb', version=__version__))
DB_DATA_DIR = pathlib.Path(appdirs.user_data_dir('opencefadb'))

USER_DATA_DIR.mkdir(parents=True, exist_ok=True)
DB_DATA_DIR.mkdir(parents=True, exist_ok=True)

GLOBAL_PACKAGE_DIR = DB_DATA_DIR



import pathlib
import unittest

import toml
import yaml

from opencefadb import __version__, get_package_meta

__this_dir__ = pathlib.Path(__file__).parent


def parse_to_py_version(vstr):
    """Parse a version string and return a standardized version string.
    e.g. removes -rc.X to rcX
    """
    return vstr.replace("-rc.", "rc")


class TestVersion(unittest.TestCase):

    def test_version(self):
        this_version = "x.x.x"
        pyproject_filename = __this_dir__ / "../pyproject.toml"
        with open(pyproject_filename, "r") as f:
            pyproject_data = toml.load(f)
            this_version = parse_to_py_version(pyproject_data["project"]["version"])
        self.assertEqual(__version__, this_version)

    def test_codemeta(self):
        """checking if the version in codemeta.json is the same as the one of the toolbox"""

        codemeta = get_package_meta()

        assert parse_to_py_version(codemeta["version"]) == __version__
        self.assertIn("@id", codemeta)
        with open(__this_dir__ / "../CITATION.cff", "r") as f:
            ccf = yaml.safe_load(f)
        self.assertIn(ccf["doi"], codemeta["@id"])

    def test_citation_cff(self):
        citation_cff = __this_dir__ / "../CITATION.cff"
        with open(citation_cff, "r") as f:
            cff = yaml.safe_load(f)
        self.assertTrue(
            "todo" not in cff["doi"].lower(), "Please replace 'todo' in CITATION.cff"
        )

    def test_citation_cff_version(self):
        """checking if the version in CITATION.cff is the same as the one of the opencefadb"""
        this_version = "x.x.x"
        setupcfg_filename = __this_dir__ / "../CITATION.cff"
        with open(setupcfg_filename, "r") as f:
            lines = f.readlines()
            for line in lines:
                if "version: " in line:
                    this_version = parse_to_py_version(line.split(":")[-1].strip())
                # elif 'date-released:' in line:
                #     # check if the date is the same as the one in codemeta.json
                #     date_str = line.split(':')[-1].strip()
                #     dt = datetime.strptime(date_str, '%Y-%m-%d')
        self.assertEqual(__version__, this_version)
        # self.assertEqual(dt.strftime('%Y-%m-%d'), datetime.now().date().strftime('%Y-%m-%d'))



import logging
import pathlib

import rdflib
from h5rdmtoolbox.catalog import InMemoryRDFStore, RDFStore, RDFStore
from rdflib.plugins.stores.sparqlstore import SPARQLStore

# from .._abstracts import OpenCeFaDBRDFStore

logger = logging.getLogger("opencefadb")


class RDFFileStore(InMemoryRDFStore):

    def reset(self, *args, **kwargs):
        self._filenames = []
        self._graphs = rdflib.Graph()
        self._combined_graph = rdflib.Graph()
        return self


class RdflibSPARQLStore(RDFStore):
    """A SPARQL endpoint RDF store using rdflib's SPARQLStore."""

    _expected_file_extensions = {".ttl", ".nt", ".rdf", ".xml", ".n3"}

    def __init__(self, endpoint_url: str):
        super().__init__()
        self._store = SPARQLStore(endpoint_url)
        self._graph = rdflib.Graph(store=self._store)

    @property
    def graph(self):
        return self._graph

    def upload_file(self, filename) -> bool:
        filename = pathlib.Path(filename).resolve().absolute()
        if not filename.exists():
            raise FileNotFoundError(f"File {filename} not found.")
        if filename.suffix not in self._expected_file_extensions:
            raise ValueError(f"File type {filename.suffix} not supported.")
        g = rdflib.Graph()
        g.parse(location=str(filename))
        self._graph += g
        return True



NUMERICAL_VARIABLE_HAS_LABEL = """
@prefix sh: <http://www.w3.org/ns/shacl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix m4i: <http://w3id.org/nfdi4ing/metadata4ing#> .
@prefix ex: <http://example.org/ns#> .  

# Every m4i:NumericalVariable must have an rdfs:label
ex:NumericalVariableMustHaveLabel
    a sh:NodeShape ;
    sh:targetClass m4i:NumericalVariable ;
    sh:property [
        sh:path rdfs:label ;
        sh:message "A m4i:NumericalVariable instances must have an rdfs:label." ;
        sh:minCount 1 ;
    ] .
"""



import pathlib
import unittest

from matplotlib import pyplot as plt
from ontolutils.ex import hdf5
from ontolutils.ex.sosa import Result, Observation, ObservationCollection
from ontolutils.namespacelib import QUDT_UNIT, QUDT_KIND
from ssnolib.m4i import NumericalVariable

from opencefadb import plotting
from opencefadb.entities import WIKIDATA_ITS_FAN_V1, WIKIDATA_FAN_OPERATING_POINT
from opencefadb.models.fan_curve import DefaultLabelResolver, SemanticFanCurve

__this_dir__ = pathlib.Path(__file__).parent

from opencefadb.models.operating_point import is_operating_point


class TestModels(unittest.TestCase):

    def setUp(self):
        self.valid_observation = Observation(
            id="https://example.org/observation/1",
            has_result=[
                Result(
                    id="https://example.org/result/vfr1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/1",
                        has_standard_name="https://example.org/standard_name/air_flow_rate",
                        hasUnit=QUDT_UNIT.M3_PER_HR
                    )
                ),
                Result(
                    id="https://example.org/result/dp1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/2",
                        has_standard_name="https://example.org/standard_name/static_pressure_difference",
                        hasUnit=QUDT_UNIT.PA,
                        has_kind_of_quantity=QUDT_KIND.StaticPressure
                    )
                ),
                Result(
                    id="https://example.org/result/n1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/3",
                        has_standard_name="https://example.org/standard_name/rotational_speed",
                        hasUnit=QUDT_UNIT.PER_MIN,
                        has_kind_of_quantity=QUDT_KIND.RotationalVelocity
                    )
                ),
            ],
        )

    def test_fan_curve(self):
        """A fan curve is made up of multiple observations, hence a ObservationCollection"""
        vfr1 = Result(
            id="https://example.org/result/vfr1",
            has_numerical_variable=NumericalVariable(
                id="https://example.org/numvar/1",
                label="Volumetric Flow Rate",
                has_standard_name="https://example.org/standard_name/air_flow_rate",
                has_numerical_value=10.0,
                units=QUDT_UNIT.M3_PER_HR,
                has_kind_of_quantity=QUDT_KIND.VolumeFlowRate
            )
        )
        dp1 = Result(
            id="https://example.org/result/dp1",
            has_numerical_variable=NumericalVariable(
                id="https://example.org/numvar/2",
                label="Static Pressure Difference",
                has_standard_name="https://example.org/standard_name/static_pressure_difference",
                has_numerical_value=80.0,
                units=QUDT_UNIT.PA,
                has_kind_of_quantity=QUDT_KIND.StaticPressure
            )
        )
        n1 = Result(
            id="https://example.org/result/n1",
            has_numerical_variable=NumericalVariable(
                id="https://example.org/numvar/3",
                label="Rotational Speed",
                has_standard_name="https://example.org/standard_name/rotational_speed",
                has_numerical_value=600.0,
                has_unit=QUDT_UNIT.REV_PER_MIN,
                has_kind_of_quantity=QUDT_KIND.RotationalVelocity
            )
        )

        vfr2 = Result(
            id="https://example.org/result/vfr2",
            has_numerical_variable=NumericalVariable(
                id="https://example.org/numvar/4",
                label="Volumetric Flow Rate",
                has_standard_name="https://example.org/standard_name/air_flow_rate",
                has_numerical_value=50.0,
                has_unit=QUDT_UNIT.M3_PER_HR,
                has_kind_of_quantity=QUDT_KIND.VolumeFlowRate
            )
        )
        dp2 = Result(
            id="https://example.org/result/dp2",
            has_numerical_variable=NumericalVariable(
                label="Static Pressure Difference",
                id="https://example.org/numvar/5",
                has_standard_name="https://example.org/standard_name/static_pressure_difference",
                has_numerical_value=30.0,
                has_unit=QUDT_UNIT.PA,
                has_kind_of_quantity=QUDT_KIND.StaticPressure
            )
        )
        n2 = Result(
            id="https://example.org/result/n2",
            has_numerical_variable=NumericalVariable(
                id="https://example.org/numvar/6",
                label="Rotational Speed",
                has_standard_name="https://example.org/standard_name/rotational_speed",
                has_numerical_value=600.0,
                has_unit=QUDT_UNIT.REV_PER_MIN,
                has_kind_of_quantity=QUDT_KIND.RotationalVelocity
            )
        )

        op1 = Observation(
            id="https://example.org/observation/1",
            hadPrimarySource=hdf5.File(
                id="https://example.org/file/1",
            ),
            has_result=[vfr1, dp1, n1]
        )
        op2 = Observation(
            id="https://example.org/observation/2",
            hadPrimarySource=hdf5.File(
                id="https://example.org/file/1",
            ),
            has_result=[vfr2, dp2, n2]
        )

        observation_collection = ObservationCollection(
            id="https://example.org/observation_collection/1",
            hasFeatureOfInterest=WIKIDATA_ITS_FAN_V1,
            hasMember=[op1, op2],
            type=WIKIDATA_FAN_OPERATING_POINT
        )
        ttl = observation_collection.serialize("ttl")

        sfc = SemanticFanCurve.from_observations(
            [op1, op2],
            id="https://example.org/observation_collection/1",
            hasFeatureOfInterest=WIKIDATA_ITS_FAN_V1,
            type=WIKIDATA_FAN_OPERATING_POINT
        )
        ttl_sfc = sfc.serialize("ttl")
        self.assertEqual(ttl_sfc, ttl)

        sfc_scaled = sfc.scale(
            n=NumericalVariable(
                has_numerical_value=1200.0,
                units=QUDT_UNIT.PER_MIN,
                has_kind_of_quantity=QUDT_KIND.RotationalVelocity
            ),
            # x="https://example.org/standard_name/air_flow_rate",
            # y="https://example.org/standard_name/static_pressure_difference",
            # n="https://example.org/standard_name/rotational_speed",
        )
        print(sfc_scaled.serialize("ttl"))

        DefaultLabelResolver.LABEL_SELECTION_ORDER = {
            "label",
            "standard_name",
        }

        if False:  # change to True to enable plotting test
            with plotting.SingleAxis(
                    scale=1.0,
                    filename="test_fan_curve.svg",
            ) as dax:
                sfc.plot(
                    x="https://example.org/standard_name/air_flow_rate",
                    y="https://example.org/standard_name/static_pressure_difference",
                    xlabel=None,
                    ylabel=None,
                    label="raw",
                    marker="o",
                    linestyle='-',
                    ax=dax.ax,
                )
                sfc_scaled.plot(
                    x="https://example.org/standard_name/air_flow_rate",
                    y="https://example.org/standard_name/static_pressure_difference",
                    xlabel=None,
                    ylabel=None,
                    label="scaled",
                    marker="+",
                    linestyle='-',
                    ax=dax.ax,
                )
                plt.legend()
                plt.tight_layout()
                plt.show()

    def test_is_operating_point(self):
        is_op = is_operating_point(self.valid_observation, verbose=True)
        self.assertTrue(is_op)

        invalid_op = Observation(
            id="https://example.org/observation/2",
            has_result=[
                Result(
                    id="https://example.org/result/vfr1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/1",
                        has_standard_name="https://example.org/standard_name/air_flow_rate",
                        hasUnit=QUDT_UNIT.M3_PER_HR
                    )
                ),
                Result(
                    id="https://example.org/result/dp1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/2",
                        has_standard_name="https://example.org/standard_name/static_pressure_difference",
                        hasUnit=QUDT_UNIT.PA
                    )
                ),
            ],
        )
        is_op_invalid = is_operating_point(invalid_op, verbose=True)
        self.assertFalse(is_op_invalid)

        another_invalid_op = Observation(
            id="https://example.org/observation/3",
            has_result=[]
        )
        is_op_another_invalid = is_operating_point(another_invalid_op, verbose=True)
        self.assertFalse(is_op_another_invalid)

        another_invalid_op = Observation(
            id="https://example.org/observation/4",
            has_result=[
                Result(
                    id="https://example.org/result/vfr1"
                ),
                Result(
                    id="https://example.org/result/dp1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/2",
                    )
                ),
                Result(
                    id="https://example.org/result/n1",
                    has_numerical_variable=NumericalVariable(
                        id="https://example.org/numvar/3",
                        has_standard_name="https://example.org/standard_name/temperature",  # wrong kind
                        has_unit=QUDT_UNIT.K
                    )
                ),
            ],
        )
        is_op_another_invalid = is_operating_point(another_invalid_op, verbose=True)
        self.assertFalse(is_op_another_invalid)



import enum
import hashlib
import logging
import pathlib
from concurrent.futures import ThreadPoolExecutor
from typing import List, Union, Optional, Dict

import requests
from rdflib import Graph, Namespace, URIRef
from tqdm import tqdm

from ._version import __version__

HDF = Namespace("http://purl.allotrope.org/ontologies/hdf5/1.8#")
SSNO = Namespace("https://matthiasprobst.github.io/ssno#")
IDX = Namespace("urn:h5cat:index#")
logger = logging.getLogger("opencefadb")
USER_AGENT_HEADER = {
    "User-Agent": f"opencefadb/{__version__} (https://github.com/matthiasprobst/opencefadb)",
}


def opencefa_print(*args, **kwargs):
    """opencefa_print is the OpenCeFa DB print function."""
    verbose = kwargs.pop("verbose", True)
    if verbose:
        print(" [OpenCeFa DB]:", *args, **kwargs)


class ExportFormat(enum.Enum):
    JSON_LD = "json-ld"
    DCAT = "dcat-ap"
    JSON = "json"
    CODEMETA = "Codemeta"
    CFF = "cff"


def _parse_checksum_algorithm(algorithm: str) -> str:
    algorithm = str(algorithm)
    if algorithm in ("sha256", "sha-256", "sha_256"):
        return "sha256"
    elif algorithm in ("md5", "md-5", "md_5"):
        return "md5"
    elif algorithm in ("sha1", "sha-1", "sha_1"):
        return "sha1"
    if "spdx.org/rdf/terms#" in algorithm:
        return algorithm.rsplit("_", 1)[-1]
    elif algorithm.startswith("http:"):
        return algorithm.rsplit('/', 1)[-1].lower()
    raise ValueError(f"Unsupported checksum algorithm: {algorithm}")


def download_file(download_url,
                  target_filename=None,
                  checksum: str = None,
                  checksum_algorithm: str = None) -> pathlib.Path:
    """Downloads from a URL"""
    if checksum is not None:
        if checksum_algorithm is None:
            raise ValueError("If checksum is provided, checksum_algorithm must also be provided.")
        checksum_algorithm = _parse_checksum_algorithm(checksum_algorithm)
        checksum = {'value': checksum, 'algorithm': checksum_algorithm}
    if target_filename is None:
        if str(download_url).endswith("/content"):
            target_filename = download_url.rsplit("/", 2)[-2]
        else:
            target_filename = download_url.split("/")[-1]
    target_filename = pathlib.Path(target_filename)
    logger.debug(f"Downloading metadata file from URL {download_url}...")
    response = requests.get(download_url, stream=True, headers=USER_AGENT_HEADER)
    response.raise_for_status()

    total_size = int(response.headers.get('content-length', 0))  # Get total file size
    chunk_size = 1024  # Define chunk size for download

    hasher = None
    if checksum is not None:
        try:
            hasher = hashlib.new(checksum_algorithm)
        except ValueError:
            raise ValueError(f"Unsupported checksum algorithm: {checksum_algorithm}")

    with tqdm(total=total_size, unit='B', unit_scale=True, desc=target_filename.name, initial=0) as progress:
        with open(target_filename, 'wb') as f:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:  # Filter out keep-alive chunks
                    f.write(chunk)
                    progress.update(len(chunk))
                    if hasher:
                        hasher.update(chunk)

    assert target_filename.exists(), f"File {target_filename} does not exist."
    logger.debug(f"Download successful.")

    if hasher:
        file_checksum = hasher.hexdigest()
        if file_checksum != checksum['value']:
            raise ValueError(
                f"Checksum mismatch for {target_filename}: expected {checksum['value']}, got {file_checksum}")
        logger.debug(f"Checksum verification successful.")

    return target_filename


def download_multiple_files(
        urls,
        target_filenames,
        max_workers=4,
        checksums: Optional[List[Dict]] = None) -> List[pathlib.Path]:
    """Download multiple files concurrently."""
    if max_workers == 1:
        return [download_file(url, target_filename, **checksum) for url, target_filename, checksum in
                zip(urls, target_filenames, checksums)]

    results = []
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [
            executor.submit(download_file, url, target_filename, checksum["checksum"], checksum["checksum_algorithm"])
            for url, target_filename, checksum in
            zip(urls, target_filenames, checksums)]
        for future in futures:
            # Wait for each future to complete (can handle exceptions if needed)
            try:
                results.append(future.result())
            except Exception as e:
                print(f"Failed to download a file: {e}")
    return results


def compute_sha256(data: str) -> str:
    """Compute the SHA256 checksum of a string."""
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def compute_md5(filename: Union[str, pathlib.Path]) -> str:
    """Compute the MD5 checksum of a file."""
    import hashlib
    filename = pathlib.Path(filename)
    if not filename.exists():
        raise FileNotFoundError(f"File {filename} does not exist.")
    hasher = hashlib.md5()
    with open(filename, 'rb') as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            hasher.update(byte_block)
    return hasher.hexdigest()


def remove_none(obj):
    if isinstance(obj, dict):
        return {k: remove_none(v) for k, v in obj.items() if v is not None}
    if isinstance(obj, list):
        return [remove_none(v) for v in obj if v is not None]
    return obj


def build_infile_index_via_parents_for_graphdb(graphdb_store, include_rootgroup: bool = True):
    index_graph_iri: str = "urn:h5cat:index"
    source_graph_iri = None
    clear_first = True

    where_graph_open = f"GRAPH <{source_graph_iri}> {{" if source_graph_iri else ""
    where_graph_close = "}" if source_graph_iri else ""

    # 1) Clear old index graph if requested
    if clear_first:
        clear_update = f"CLEAR GRAPH <{index_graph_iri}>"
        graphdb_store._post_update(clear_update)

    # 2) Build INSERT ... WHERE ...
    construct_triples = "?dataset idx:inFile ?file .\n"
    if include_rootgroup:
        construct_triples += "?dataset idx:rootGroup ?root .\n"

    insert_update = f"""
    PREFIX hdf:  <http://purl.allotrope.org/ontologies/hdf5/1.8#>
    PREFIX ssno: <https://matthiasprobst.github.io/ssno#>
    PREFIX idx:  <urn:h5cat:index#>

    INSERT {{
      GRAPH <{index_graph_iri}> {{
        {construct_triples}
      }}
    }}
    WHERE {{
      {where_graph_open}

      ?file a hdf:File ;
            hdf:rootGroup ?root .

      # Any node reachable from the root that has ssno:hasStandardName is a dataset in your model
      ?root hdf:member+ ?dataset .
      ?dataset ssno:hasStandardName ?sn .

      {where_graph_close}
    }}
    """
    graphdb_store._post_update(insert_update)


def build_infile_index_via_parents_for_graph(g: Graph, include_rootgroup: bool = True) -> Graph:
    idx_g = Graph()
    idx_g.bind("idx", IDX)

    # rootGroup -> file
    root_to_file = {root: file_uri for file_uri, root in g.subject_objects(HDF.rootGroup)}

    # candidates from either direction of hasStandardName
    candidates = set(g.subjects(SSNO.hasStandardName, None)) | set(g.objects(None, SSNO.hasStandardName))

    # keep only nodes that look "contained": have a parent via hdf:member or are a rootGroup
    datasets = {
        n for n in candidates
        if (n in root_to_file) or any(True for _ in g.subjects(HDF.member, n))
    }

    memo: dict[URIRef, tuple[URIRef, URIRef]] = {}

    def resolve_to_root_and_file(start: URIRef) -> tuple[URIRef, URIRef]:
        path = []
        node = start
        seen = set()

        while True:
            if node in memo:
                root, file_uri = memo[node]
                break

            if node in root_to_file:
                root = node
                file_uri = root_to_file[node]
                break

            if node in seen:
                raise ValueError(f"Cycle detected in hdf:member chain at {node}")
            seen.add(node)
            path.append(node)

            parents = list(g.subjects(HDF.member, node))
            if not parents:
                raise KeyError(f"No parent found via hdf:member for node {node}")
            node = parents[0]  # you said it's unique per dataset/file

        for n in path:
            memo[n] = (root, file_uri)
        return root, file_uri

    for ds in datasets:
        root, file_uri = resolve_to_root_and_file(ds)
        idx_g.add((ds, IDX.inFile, file_uri))
        if include_rootgroup:
            idx_g.add((ds, IDX.rootGroup, root))

    return idx_g
